from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class CompositeWriteNumber(Component):
    """
    Writes up to 32 numbers to a composite link in a single logic tick. Only connected channels will be modified.
    The number of inputs and channel to begin writing at can be configured.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        start_channel_property: int = 1,
        channel_count_property: int = 1,
        composite_signal_input: Optional[Input] = None,
        start_channel_input: Optional[Input] = None,
        channel_inputs: Optional[dict[int, Input]] = None,
    ):
        super().__init__(ComponentType.CompositeWriteNumber, component_id, position)
        self.start_channel_property: int = start_channel_property
        self.channel_count_property: int = channel_count_property
        self.composite_signal_input: Optional[Input] = composite_signal_input
        self.start_channel_input: Optional[Input] = start_channel_input
        self.channel_inputs: dict[int, Input] = channel_inputs or {}

    @property
    def height(self) -> float:
        return (
            0.5
            + self.channel_count_property * 0.25
            + (0.25 if self.start_channel_property == 0 else 0)
        )

    @staticmethod
    def from_xml(element: XMLParserElement) -> CompositeWriteNumber:
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, _ = Component._basic_in_parsing(obj)
        start_channel_property: int = int(obj.attributes.get("offset", "0")) + 1
        channel_count_property: int = int(
            obj.attributes.get("count", "32")
        )  # yes, absent means 32
        channel_inputs: dict[int, Input] = {
            i: current_input
            for i in range(1, 33)
            if (current_input := inputs.get(str(i)))
        }
        return CompositeWriteNumber(
            component_id,
            position,
            start_channel_property,
            channel_count_property,
            inputs.get("c"),
            inputs.get("off"),
            channel_inputs,
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        attributes: dict[str, str] = {}
        if self.start_channel_property != 1:
            attributes["offset"] = str(self.start_channel_property - 1)
        if self.channel_count_property != 32:
            attributes["count"] = str(self.channel_count_property)
        inputs: list[Optional[Input]] = [Input(0)] * self.channel_count_property
        for i, current_input in self.channel_inputs.items():
            inputs[i - 1] = current_input
        children: list[XMLParserElement] = self._pos_in_to_xml(
            *inputs,
            named_inputs={
                "c": self.composite_signal_input,
                "off": self.start_channel_input,
            },
        )
        return attributes, children



from __future__ import annotations

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.NumberProperty import NumberInput, NumberProperty
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class PropertySlider(Component):
    """
    Adds a custom slider that will be seen on the microcontroller's property panel when placed on a vehicle.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        min_property: NumberInput = None,
        max_property: NumberInput = None,
        value_property: NumberInput = None,
        rounding_property: NumberInput = None,
        name: str = "value",
    ):
        super().__init__(ComponentType.PropertySlider, component_id, position)
        self.rounding_property: NumberProperty = NumberProperty.from_input(
            rounding_property, "int"
        )
        self.min_property: NumberProperty = NumberProperty.from_input(
            min_property, "min"
        )
        self.max_property: NumberProperty = NumberProperty.from_input(
            max_property, "max"
        )
        self.value_property: NumberProperty = NumberProperty.from_input(
            value_property, "v"
        )
        self.name: str = name

    @property
    def height(self) -> float:
        return 0.5

    @staticmethod
    def from_xml(element: XMLParserElement) -> PropertySlider:
        assert element.tag == "c", f"invalid PropertySlider {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.PropertySlider.value
        ), f"Not an PropertySlider {element}"
        obj: XMLParserElement = element.children[0]
        component_id, position, _, properties = PropertySlider._basic_in_parsing(obj)
        return PropertySlider(
            component_id,
            position,
            properties.get("min"),
            properties.get("max"),
            properties.get("v"),
            properties.get("int"),
            obj.attributes.get("name", "value"),
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        return {"name": self.name}, self._pos_in_to_xml(
            properties={
                "min": self.min_property,
                "max": self.max_property,
                "v": self.value_property,
                "int": self.rounding_property,
            }
        )



from enum import Enum


class _Enum(Enum):
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}.{self.name}"


class SignalType(_Enum):
    """
    Signal types
    """

    Boolean = 0
    Number = 1
    Composite = 5
    Video = 6
    Audio = 7


class NodeMode(_Enum):
    """
    Modes of nodes (in/out connections)
    """

    Output = 0
    Input = 1


class TooltipMode(_Enum):
    """
    Modes of Tooltips, depending on the error/signal input
    """

    Always = 0
    IfOn = 1
    IfOff = 2


class PulseMode(_Enum):
    """
    Modes of the Pulse Component, note that OffToOn is the default mode
    """

    OnToOff = 0
    OffToOn = 1
    Always = 2


class TimerUnit(_Enum):
    """
    Units of timer Components
    """

    Seconds = 0
    Ticks = 1


class ComponentType(_Enum):
    """
    Types of Components and their respective ids
    """

    NOT = 0
    AND = 1
    OR = 2
    XOR = 3
    NAND = 4
    NOR = 5
    Add = 6
    Subtract = 7
    Multiply = 8
    Divide = 9
    ArithmeticFunction3In = 10
    Clamp = 11
    Threshold = 12
    MemoryRegister = 13
    Abs = 14
    ConstantNumber = 15
    ConstantOn = 16
    GreaterThan = 17
    LessThan = 18
    PropertySlider = 19
    PropertyDropdown = 20
    NumericalJunction = 21
    NumericalSwitchbox = 22
    PIDController = 23
    SRLatch = 24
    JKFlipFlop = 25
    Capacitor = 26
    Blinker = 27
    PushToToggle = 28
    CompositeReadBoolean = 29
    CompositeWriteBooleanDeprecated = 30
    CompositeReadNumber = 31
    CompositeWriteNumberDeprecated = 32
    PropertyToggle = 33
    PropertyNumber = 34
    Delta = 35
    ArithmeticFunction8In = 36
    UpDownCounter = 37
    Modulo = 38
    PIDControllerAdvanced = 39
    CompositeWriteNumber = 40
    CompositeWriteBoolean = 41
    Equal = 42
    TooltipNumber = 43
    TooltipBoolean = 44
    ArithmeticFunction1In = 45
    BooleanFunction4In = 46
    BooleanFunction8In = 47
    Pulse = 48
    TimerTON = 49
    TimerTOF = 50
    TimerRTO = 51
    TimerRTF = 52
    CompositeSwitchbox = 53
    NumberToCompositeBinary = 54
    CompositeBinaryToNumber = 55
    LuaScript = 56
    VideoSwitchbox = 57
    PropertyText = 58
    AudioSwitchbox = 59



from pathlib import Path

from sw_mc_lib import Microcontroller, XMLFormatter
from sw_mc_lib.XMLParser import XMLParser


def read_mc_from_file(path: Path) -> Microcontroller:
    """
    Read a Microcontroller from a file

    :param path: The path to the file
    :return: The Microcontroller
    """
    with path.open("r") as file:
        content = file.read()
    element = XMLParser(content).parse()
    return Microcontroller.from_xml(element)


def write_mc_to_file(mc: Microcontroller, path: Path) -> None:
    """
    Write a Microcontroller to a file

    :param mc: The Microcontroller to write
    :param path: The path to the file
    :return: None
    """
    element = mc.to_xml()
    formatted = XMLFormatter.format(element, header=True)
    with path.open("w") as file:
        file.write(formatted)



from inspect import Signature, signature
from typing import Any, Optional


def generic_str(
    self: Any,
    ignored_keywords: Optional[list[str]] = None,
    explicit_keywords: Optional[list[str]] = None,
    newlines: bool = True,
) -> str:
    """
    Return a representation of a class that is valid python, and optionally also indented.

    :param self: The class that needs a representation
    :param ignored_keywords: Ignore these attributes
    :param explicit_keywords: Include these attributes, regardless of ignores and signatures
    :param newlines: use newlines and indentation
    :return: The representation
    """
    newline_or_space: str = "\n" if newlines else " "
    newline_or_nothing: str = "\n" if newlines else ""
    string: str = f"{self.__class__.__name__}({newline_or_nothing}"
    dir: list[str] = []
    self_signature: Signature = signature(self.__init__)
    ignored_keywords = (ignored_keywords or []) + ["self"]
    for i, content in self_signature.parameters.items():
        if i not in ignored_keywords and getattr(self, i, None) != content.default:
            dir.append(i)
    if explicit_keywords:
        for i in explicit_keywords:
            if i not in dir:
                dir.append(i)
    if dir:
        elements = [f"{i}={getattr(self, i)!r}" for i in dir]
        if all("\n" not in i for i in elements) and sum(len(i) for i in elements) < 80:
            return f"{self.__class__.__name__}({', '.join(elements)})"
        for i in dir:
            string += f'{"    " if newlines else ""}{i}='
            lines = repr(getattr(self, i)).splitlines()
            string += f"{lines.pop(0)}{newline_or_space}"
            for j in lines:
                string += f'{"    "}{j}{newline_or_space}'
            string = f"{string[:-1]},{newline_or_space}"
        string = string[:-1]
    else:
        return f"{self.__class__.__name__}()"
    return f"{string[:-1]}{newline_or_nothing})"



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.NumberProperty import NumberInput, NumberProperty
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class UpDownCounter(Component):
    """
    Has an internal value that will increase and decrease when receiving different signals.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        up_input: Optional[Input] = None,
        down_input: Optional[Input] = None,
        reset_input: Optional[Input] = None,
        min_property: NumberInput = None,
        max_property: NumberInput = None,
        reset_property: NumberInput = None,
        increment_property: NumberInput = None,
        clamp: bool = False,
    ):
        super().__init__(ComponentType.UpDownCounter, component_id, position)
        self.up_input: Optional[Input] = up_input
        self.down_input: Optional[Input] = down_input
        self.reset_input: Optional[Input] = reset_input
        self.increment_property: NumberProperty = NumberProperty.from_input(
            increment_property, "i"
        )
        self.min_property: NumberProperty = NumberProperty.from_input(
            min_property, "min"
        )
        self.max_property: NumberProperty = NumberProperty.from_input(
            max_property, "max"
        )
        self.reset_property: NumberProperty = NumberProperty.from_input(
            reset_property, "r"
        )
        self.clamp: bool = clamp

    @property
    def height(self) -> float:
        return 1.0

    @staticmethod
    def from_xml(element: XMLParserElement) -> UpDownCounter:
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, properties = Component._basic_in_parsing(obj)
        return UpDownCounter(
            component_id,
            position,
            inputs.get("1"),
            inputs.get("2"),
            inputs.get("3"),
            properties.get("min"),
            properties.get("max"),
            properties.get("r"),
            properties.get("i"),
            obj.attributes.get("m", "0") == "1",
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        return {"m": str(int(self.clamp))}, self._pos_in_to_xml(
            self.up_input,
            self.down_input,
            self.reset_input,
            properties={
                "min": self.min_property,
                "max": self.max_property,
                "r": self.reset_property,
                "i": self.increment_property,
            },
        )



from __future__ import annotations

from itertools import chain

from .Component import Component
from .Node import Node
from .XMLElement import XMLElement, XMLParserElement


class Microcontroller(XMLElement):
    """
    A Stormworks Microcontroller
    """

    def __init__(
        self,
        name: str,
        description: str,
        width: int,
        length: int,
        nodes: list[Node],
        components: list[Component],
    ):
        self.name: str = name
        self.description: str = description
        self.width: int = width
        self.length: int = length
        self.nodes: list[Node] = nodes
        self.components: list[Component] = components

    @staticmethod
    def from_xml(element: XMLParserElement) -> Microcontroller:
        # pylint: disable=too-many-locals
        assert element.tag in ("microprocessor", "microprocessor_definition")
        name: str = element.attributes.get("name", "")
        description: str = element.attributes.get("description", "")
        width: int = int(element.attributes.get("width", "0"))
        length: int = int(element.attributes.get("length", "0"))
        nodes: list[Node] = []
        nodes_elem: XMLParserElement = element.children[0]
        assert nodes_elem.tag == "nodes"
        for node in nodes_elem.children:
            nodes.append(Node.from_xml(node))
        group_elem: XMLParserElement = element.children[1]
        assert group_elem.tag == "group"
        components_elem: XMLParserElement = group_elem.children[1]
        if components_elem.tag == "data":
            group_elem.children.pop(1)
            components_elem = group_elem.children[1]
        assert components_elem.tag == "components"
        components: list[Component] = []
        for component in components_elem.children:
            components.append(Component.from_xml(component))
        component_bridge_elem: XMLParserElement = group_elem.children[2]
        assert component_bridge_elem.tag == "components_bridge"
        for child in component_bridge_elem.children:
            component_id = int(child.children[0].attributes.get("id", "0"))
            for xml_node in nodes:
                if xml_node.component_id == component_id:
                    xml_node.add_component_bridge(child)
                    break
        return Microcontroller(name, description, width, length, nodes, components)

    def to_xml(self) -> XMLParserElement:
        nodes_elem: XMLParserElement = XMLParserElement("nodes")
        for node in self.nodes:
            nodes_elem.children.append(node.to_xml())
        group_elem: XMLParserElement = XMLParserElement("group")
        group_elem.children.append(
            XMLParserElement(
                "data", {}, [XMLParserElement("inputs"), XMLParserElement("outputs")]
            )
        )
        components_elem: XMLParserElement = XMLParserElement("components")
        for component in self.components:
            components_elem.children.append(component.to_xml())
        group_elem.children.append(components_elem)
        component_bridge_elem: XMLParserElement = XMLParserElement("components_bridge")
        for node in self.nodes:
            component_bridge_elem.children.append(node.to_component_bridge())
        group_elem.children.append(component_bridge_elem)
        group_elem.children.append(XMLParserElement("groups"))
        components_states_elem: XMLParserElement = XMLParserElement("component_states")
        for i, component in enumerate(self.components):
            components_states_elem.children.append(component.to_state_xml(i))
        group_elem.children.append(components_states_elem)
        component_bridge_states_elem: XMLParserElement = XMLParserElement(
            "component_bridge_states"
        )
        for i, node in enumerate(self.nodes):
            component_bridge_states_elem.children.append(
                node.to_component_bridge_state(i)
            )
        group_elem.children.append(component_bridge_states_elem)
        group_elem.children.append(XMLParserElement("group_states"))
        attributes: dict[str, str] = {
            "name": self.name,
            "description": self.description,
            "width": str(self.width),
            "length": str(self.length),
            "id_counter": str(self.id_counter),
            "id_counter_node": str(self.node_counter),
        }
        return XMLParserElement("microprocessor", attributes, [nodes_elem, group_elem])

    def add_new_component(self, component: Component) -> None:
        """
        Adds a new :class:`Component <sw_mc_lib.Component.Component>`. Sets the right `component_id`.

        :param component: The component to add, `component_id` may be any value
        :return: None
        """
        component.component_id = self.id_counter + 1
        self.components.append(component)

    def add_new_node(self, node: Node) -> None:
        """
        Adds a new :class:`Node <sw_mc_lib.Node.Node>`. Sets the right `component_id` and `node_id`.

        :param node: The Node to add
        :return: None
        """
        node.component_id = self.id_counter + 1
        node.node_id = self.node_counter + 1
        self.nodes.append(node)

    @property
    def id_counter(self) -> int:
        """
        Stormworks internal counter of what the largest component_id is.

        :return: The currently largest component_id
        """
        components = (comp.component_id for comp in self.components)
        nodes = (node.component_id for node in self.nodes)
        return max(chain(components, nodes, (0,)))

    @property
    def node_counter(self) -> int:
        """
        Stormworks internal counter of what the largest node_id is.

        :return: The currently largest node_id
        """
        nodes = (node.node_id for node in self.nodes)
        return max(chain(nodes, (0,)))

    def clone(self) -> Microcontroller:
        """
        Creates a deep copy of this microcontroller.

        :return: A deep copy of this microcontroller.
        """
        return Microcontroller(
            self.name,
            self.description,
            self.width,
            self.length,
            [node.clone() for node in self.nodes],
            [component.clone() for component in self.components],
        )



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.NumberProperty import NumberInput, NumberProperty
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class Threshold(Component):
    """
    Outputs an on/off signal indicating whether or not the input value is within a set threshold.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        min_property: NumberInput = None,
        max_property: NumberInput = None,
        number_input: Optional[Input] = None,
    ):
        super().__init__(ComponentType.Threshold, component_id, position)
        self.number_input: Optional[Input] = number_input
        self.min_property: NumberProperty = NumberProperty.from_input(
            min_property, "min"
        )
        self.max_property: NumberProperty = NumberProperty.from_input(
            max_property, "max"
        )

    @property
    def height(self) -> float:
        return 0.5

    @staticmethod
    def from_xml(element: XMLParserElement) -> Threshold:
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, properties = Component._basic_in_parsing(obj)
        return Threshold(
            component_id,
            position,
            properties.get("min"),
            properties.get("max"),
            inputs.get("1"),
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        return {}, self._pos_in_to_xml(
            self.number_input,
            properties={"min": self.min_property, "max": self.max_property},
        )



from .abs import Abs
from .add import Add
from .and_ import AND
from .arithmetic_function_1_in import ArithmeticFunction1In
from .arithmetic_function_3_in import ArithmeticFunction3In
from .arithmetic_function_8_in import ArithmeticFunction8In
from .audio_switchbox import AudioSwitchbox
from .blinker import Blinker
from .boolean_function_4_in import BooleanFunction4In
from .boolean_function_8_in import BooleanFunction8In
from .capacitor import Capacitor
from .clamp import Clamp
from .composite_binary_to_number import CompositeBinaryToNumber
from .composite_read_boolean import CompositeReadBoolean
from .composite_read_number import CompositeReadNumber
from .composite_switchbox import CompositeSwitchbox
from .composite_write_boolean import CompositeWriteBoolean
from .composite_write_number import CompositeWriteNumber
from .constant_number import ConstantNumber
from .constant_on import ConstantOn
from .delta import Delta
from .divide import Divide
from .equal import Equal
from .greater_than import GreaterThan
from .jk_flip_flop import JKFlipFlop
from .less_than import LessThan
from .lua_script import LuaScript
from .memory_register import MemoryRegister
from .modulo import Modulo
from .multiply import Multiply
from .nand import NAND
from .nor import NOR
from .not_ import NOT
from .number_to_composite_binary import NumberToCompositeBinary
from .numerical_junction import NumericalJunction
from .numerical_switchbox import NumericalSwitchbox
from .or_ import OR
from .pid_controller import PIDController
from .pid_controller_advanced import PIDControllerAdvanced
from .property_dropdown import PropertyDropdown
from .property_number import PropertyNumber
from .property_slider import PropertySlider
from .property_text import PropertyText
from .property_toggle import PropertyToggle
from .pulse import Pulse
from .push_to_toggle import PushToToggle
from .sr_latch import SRLatch
from .subtract import Subtract
from .threshold import Threshold
from .timer_rtf import TimerRTF
from .timer_rto import TimerRTO
from .timer_tof import TimerTOF
from .timer_ton import TimerTON
from .tooltip_boolean import TooltipBoolean
from .tooltip_number import TooltipNumber
from .up_down_counter import UpDownCounter
from .video_switchbox import VideoSwitchbox
from .xor import XOR

__all__ = [
    "Abs",
    "Add",
    "AND",
    "ArithmeticFunction1In",
    "ArithmeticFunction3In",
    "ArithmeticFunction8In",
    "AudioSwitchbox",
    "Blinker",
    "BooleanFunction4In",
    "BooleanFunction8In",
    "Capacitor",
    "Clamp",
    "CompositeBinaryToNumber",
    "CompositeReadBoolean",
    "CompositeReadNumber",
    "CompositeSwitchbox",
    "CompositeWriteBoolean",
    "CompositeWriteNumber",
    "ConstantNumber",
    "ConstantOn",
    "Delta",
    "Divide",
    "Equal",
    "GreaterThan",
    "JKFlipFlop",
    "LessThan",
    "LuaScript",
    "MemoryRegister",
    "Modulo",
    "Multiply",
    "NAND",
    "NOR",
    "NOT",
    "NumberToCompositeBinary",
    "NumericalJunction",
    "NumericalSwitchbox",
    "OR",
    "PIDController",
    "PIDControllerAdvanced",
    "PropertyDropdown",
    "PropertyNumber",
    "PropertySlider",
    "PropertyText",
    "PropertyToggle",
    "Pulse",
    "PushToToggle",
    "SRLatch",
    "Subtract",
    "Threshold",
    "TimerRTF",
    "TimerRTO",
    "TimerTOF",
    "TimerTON",
    "TooltipBoolean",
    "TooltipNumber",
    "UpDownCounter",
    "VideoSwitchbox",
    "XOR",
]



import sw_mc_lib.Types

from .Component import Component
from .Input import Input
from .Microcontroller import Microcontroller
from .Node import Node, NodePosition
from .NumberProperty import NumberProperty
from .Position import Position
from .XMLFormatter import XMLParserElement, format
from .XMLParser import parse



from typing import TypedDict

from sw_mc_lib import Component, Microcontroller
from sw_mc_lib import Node as MCNode
from sw_mc_lib import Position

NodeID = int


class Node(TypedDict):
    """A node in the graph for layout purposes."""

    id: NodeID
    inputs: list[NodeID]
    width: float
    height: float


def _layout_nodes_variable_heights(
    nodes: list[Node],
    width: float = 1.0,
    spacing: float = 0.25,
) -> dict[NodeID, Position]:
    """
    A practical implementation of the Sugiyama-style layered layout for nodes with variable heights.

    - Nodes are assumed to be horizontal boxes: inputs on left, outputs on right.
    - `nodes` is a list of Node typed dicts where `inputs` are NodeIDs pointing to upstream nodes.
    - Returns a mapping NodeID -> Position.

    Algorithm outline:
    1. Build directed graph edges input -> node.
    2. Condense strongly-connected components (Tarjan) to handle cycles.
    3. Assign layers using longest-path layering on the component DAG (sources at layer 0).
    4. Reduce crossings by repeated barycenter ordering between adjacent layers.
    5. Assign x coordinates by layer index and y coordinates by stacking nodes in each layer considering node heights and spacing.

    This is intended as a robust, readable implementation rather than the most optimized.
    """
    # pylint: disable=too-many-locals,too-many-statements,too-many-branches

    # --- Helpers ---
    node_map: dict[NodeID, Node] = {n["id"]: n for n in nodes}

    # Build adjacency: edges input -> node
    adj: dict[NodeID, list[NodeID]] = {nid: [] for nid in node_map}
    rev_adj: dict[NodeID, list[NodeID]] = {nid: [] for nid in node_map}
    for n in nodes:
        for src in n["inputs"]:
            if src not in node_map:
                # ignore missing references
                continue
            adj[src].append(n["id"])  # src -> n
            rev_adj[n["id"]].append(src)

    # --- Break cycles by removing edges (heuristic) ---
    # Instead of condensing SCCs, iteratively find a cycle and remove a single edge from that
    # cycle. Prefer removing the edge whose *source* node has the largest outbound degree
    # (heuristic: removing from hub-like nodes minimizes loss of connectivity overall).
    # We operate on a copy of `adj` so the original adjacency is still available for later use
    # (for example to draw or inspect removed edges).
    adj_copy: dict[NodeID, list[NodeID]] = {nid: list(adj[nid]) for nid in adj}

    def _find_cycle_in_copy() -> list[NodeID] | None:
        visited: set[NodeID] = set()
        stack: list[NodeID] = []
        onstack: set[NodeID] = set()

        def dfs(u: NodeID) -> list[NodeID] | None:
            visited.add(u)
            stack.append(u)
            onstack.add(u)
            for v in adj_copy.get(u, []):
                if v not in visited:
                    res = dfs(v)
                    if res:
                        return res
                elif v in onstack:
                    # found a cycle: return the cycle nodes (v...u, and back to v)
                    try:
                        idx = stack.index(v)
                        return stack[idx:] + [v]
                    except ValueError:
                        return [v, u, v]
            stack.pop()
            onstack.remove(u)
            return None

        for n in node_map:
            if n not in visited:
                c = dfs(n)
                if c:
                    return c
        return None

    # Greedily remove edges until the graph is acyclic
    while True:
        cycle = _find_cycle_in_copy()
        if not cycle:
            break
        # cycle is a list like [n1, n2, ..., nk, n1]
        # choose an edge (s->t) in that cycle whose source s has the largest out-degree
        best_edge: tuple[NodeID, NodeID] | None = None
        best_outdeg = -1
        for i in range(len(cycle) - 1):
            s = cycle[i]
            t = cycle[i + 1]
            outdeg = len(adj_copy.get(s, []))
            if outdeg > best_outdeg:
                best_outdeg = outdeg
                best_edge = (s, t)
        if best_edge:
            s, t = best_edge
            if t in adj_copy.get(s, []):
                adj_copy[s].remove(t)

    # Now adj_copy is a DAG (acyclic). Build reverse adjacency for barycenter and layering.
    dag_adj: dict[NodeID, list[NodeID]] = adj_copy
    dag_rev: dict[NodeID, list[NodeID]] = {nid: [] for nid in node_map}
    for u, outs in dag_adj.items():
        for v in outs:
            dag_rev[v].append(u)

    # --- Longest-path layering on component DAG ---
    # topological order via Kahn
    indeg = {i: 0 for i in dag_adj}
    for u, outs in dag_adj.items():
        for v in outs:
            indeg[v] += 1
    queue = [u for u, d in indeg.items() if d == 0]
    topo: list[int] = []
    qidx = 0
    while qidx < len(queue):
        u = queue[qidx]
        qidx += 1
        topo.append(u)
        for v in dag_adj[u]:
            indeg[v] -= 1
            if indeg[v] == 0:
                queue.append(v)

    if len(topo) != len(dag_adj):
        # Shouldn't happen after SCC condensation, but guard just in case.
        topo = list(dag_adj.keys())

    comp_layer: dict[int, int] = {i: 0 for i in dag_adj}
    for u in reversed(topo):
        for v in dag_rev[u]:
            comp_layer[v] = min(comp_layer[v], comp_layer[u] - 1)

    min_layer = min(comp_layer.values(), default=0)

    for u in topo:
        if dag_rev[u]:
            comp_layer[u] = max(comp_layer[v] + 1 for v in dag_rev[u])

    # Map nodes to layers
    layer_of: dict[NodeID, int] = {nid: comp_layer[nid] - min_layer for nid in node_map}

    # Build layers list (ordered)
    max_layer = max(layer_of.values()) if layer_of else 0
    layers: dict[int, list[NodeID]] = {i: [] for i in range(max_layer + 1)}
    for nid in node_map:
        layers[layer_of[nid]].append(nid)

    # Initial ordering: sort by number of inputs (heuristic) for stability
    for layer in layers:
        layers[layer].sort(key=lambda nid: (len(rev_adj[nid]), node_map[nid]["id"]))

    # --- Crossing reduction: barycenter heuristic ---
    def barycenter_order(layer_index: int, use_prev: bool) -> None:
        """Reorder nodes in layer `layer_index` using barycenters to adjacent layer.
        If use_prev is True, compute barycenters using nodes in previous layer (parents).
        Otherwise use next layer (children).
        """
        nodes_in_layer = layers[layer_index]
        if not nodes_in_layer:
            return
        neighbor_layer_index = layer_index - 1 if use_prev else layer_index + 1
        if neighbor_layer_index < 0 or neighbor_layer_index > max_layer:
            return
        neighbor_order = {nid: i for i, nid in enumerate(layers[neighbor_layer_index])}
        bary: list[tuple[float, int, NodeID]] = []  # (barycenter, original_index, nid)
        for i, nid in enumerate(nodes_in_layer):
            neighbors = rev_adj[nid] if use_prev else adj[nid]
            idxs = []
            for n in neighbors:
                if n in neighbor_order:
                    relative: int
                    if use_prev:
                        node = node_map[nid]
                        relative = node["inputs"].index(n)
                    else:
                        node = node_map[n]
                        relative = -node["inputs"].index(nid)
                    idxs.append(
                        neighbor_order[n] + (relative * 0.02)
                    )  # slight offset for input order
            if idxs:
                b = sum(idxs) / len(idxs)
                bary.append((b, i, nid))
            else:
                bary.append((float("inf"), i, nid))
        bary.sort(key=lambda t: (t[0], t[1]))
        layers[layer_index] = [t[2] for t in bary]

    # iterate barycenter passes
    passes = 24
    for _ in range(passes):
        # forward pass (left to right): use previous layer parents
        for li in range(1, max_layer + 1):
            barycenter_order(li, use_prev=True)
        # backward pass (right to left): use next layer children
        for li in range(max_layer - 1, -1, -1):
            barycenter_order(li, use_prev=False)

    # --- Coordinate assignment ---
    positions: dict[NodeID, Position] = {}

    x_offset = (max_layer * (width + spacing)) / 8 * 4
    # x per layer
    x_of_layer = {i: i * (width + spacing) - x_offset for i in range(max_layer + 1)}

    # For each layer, compute y positions by stacking nodes with spacing, centered at y=0
    for li in range(max_layer + 1):
        order = layers[li]
        if not order:
            continue
        heights = [node_map[nid]["height"] for nid in order]
        total_h = (sum(heights) + spacing * (len(order) - 1)) * 4 // 4
        start_y = -total_h / 2.0
        y = start_y
        for nid, h in zip(order, heights):
            positions[nid] = Position(x_of_layer[li], y)
            y += h + spacing

    return positions


def layout_mc(mc: Microcontroller) -> None:
    """
    Layout the components and nodes of a Microcontroller in a Sugiyama-style layered layout.
    Modifies the `position` attribute of each component and node in place.
    """
    all_components: dict[NodeID, MCNode | Component] = {}
    nodes: list[Node] = []
    for n in mc.nodes:
        node_inputs: list[NodeID] = []
        if n.input:
            node_inputs.append(n.input.component_id)
        nodes.append(
            {
                "id": n.component_id,
                "inputs": node_inputs,
                "width": 1.0,
                "height": 0.5,
            }
        )
        all_components[n.component_id] = n
    for component in mc.components:
        nodes.append(
            {
                "id": component.component_id,
                "inputs": [input.component_id for input in component.inputs],
                "width": 1.0,
                "height": component.height,
            }
        )
        if component.component_id in nodes[-1]["inputs"]:
            # There is no use of self references for layout purposes.
            nodes[-1]["inputs"].remove(component.component_id)
        all_components[component.component_id] = component
    positions = _layout_nodes_variable_heights(nodes, width=1.0, spacing=0.25)
    for nid, pos in positions.items():
        all_components[nid].position = pos



from sw_mc_lib import Component, Microcontroller
from sw_mc_lib.Components import (
    ArithmeticFunction1In,
    ArithmeticFunction3In,
    ArithmeticFunction8In,
    BooleanFunction4In,
    BooleanFunction8In,
    CompositeWriteBoolean,
    CompositeWriteNumber,
    PropertyDropdown,
    PropertyNumber,
    PropertySlider,
    PropertyText,
    PropertyToggle,
    TooltipBoolean,
    TooltipNumber,
)
from sw_mc_lib.Types import ComponentType


def remove_unused(mc: Microcontroller) -> None:
    """
    Remove all components that do not contribute to any output or global variable.

    Also checks for components that are only cyclically connected and removes them.
    :param mc: Microcontroller to optimize
    :return: None
    """
    components_by_id = {}
    to_remove: set[int] = set()
    for component in mc.components:
        if isinstance(component, (TooltipBoolean, TooltipNumber)):
            # Tooltips are outputs out of the mc of their own
            if component.display_number_input:
                to_remove.add(component.display_number_input.component_id)
        elif not isinstance(
            component,
            (
                PropertyNumber,
                PropertyDropdown,
                PropertySlider,
                PropertyText,
                PropertyToggle,
            ),
        ):
            # Properties can be accessed without any connection in lua
            components_by_id[component.component_id] = component

    for node in mc.nodes:
        if node.input:
            to_remove.add(node.input.component_id)

    while to_remove:
        current = to_remove.pop()
        if current not in components_by_id:
            continue
        current_component = components_by_id.pop(current)
        for input in current_component.inputs:
            to_remove.add(input.component_id)

    for component in components_by_id.values():
        mc.components.remove(component)


def optimize_composite_writes(mc: Microcontroller) -> None:
    """
    Optimize composite writes by setting start_channel and channel_count.

    For example, if a composite has two inputs A and B, and both are set to the same value,
    the second write is redundant and can be removed.

    :param mc: Microcontroller to optimize
    :return: None
    """
    for component in mc.components:
        if not isinstance(component, (CompositeWriteBoolean, CompositeWriteNumber)):
            continue
        max_channel: int = max(component.channel_inputs.keys(), default=0)
        min_channel: int = min(component.channel_inputs.keys(), default=1)
        channel_count: int = max_channel - min_channel + 1
        component.channel_count_property = min(
            component.channel_count_property, channel_count
        )
        start_channel: int = component.start_channel_property
        if start_channel != 0 and min_channel > 1:
            diff: int = min_channel - 1
            keys = list(component.channel_inputs.keys())
            for channel_id in keys:
                channel = component.channel_inputs[channel_id]
                component.channel_inputs[channel_id - diff] = channel
                del component.channel_inputs[channel_id]
                channel.index = str(channel_id - diff)
            component.start_channel_property = start_channel + diff


def optimize_functions(mc: Microcontroller) -> None:
    """
    Optimize the microcontroller by making function blocks as small as possible.

    :param mc: Microcontroller to optimize
    :return: None
    """
    for component in mc.components:
        if isinstance(component, ArithmeticFunction8In):
            if (
                component.a_input is None
                and component.b_input is None
                and component.c_input is None
                and component.d_input is None
                and component.w_input is None
            ):
                if component.y_input is None and component.z_input is None:
                    component.type = ComponentType.ArithmeticFunction1In
                    component.__class__ = ArithmeticFunction1In  # type: ignore
                else:
                    component.type = ComponentType.ArithmeticFunction3In
                    component.__class__ = ArithmeticFunction3In  # type: ignore
        elif isinstance(component, ArithmeticFunction3In):
            if component.y_input is None and component.z_input is None:
                component.type = ComponentType.ArithmeticFunction1In
                component.__class__ = ArithmeticFunction1In  # type: ignore
        elif isinstance(component, BooleanFunction8In):
            if (
                component.a_input is None
                and component.b_input is None
                and component.c_input is None
                and component.d_input is None
            ):
                component.type = ComponentType.BooleanFunction4In
                component.__class__ = BooleanFunction4In  # type: ignore


def optimize_tree(mc: Microcontroller) -> None:
    """
    Optimize the microcontroller by removing unused components and optimizing functions and composite writes.

    :param mc: Microcontroller to optimize
    :return: None
    """
    found: bool = True
    while found:
        found = False
        known_components: dict[Component, int] = {}
        replacements: dict[int, int] = {}
        replaced: list[int] = []
        for i, component in enumerate(mc.components):
            if component in known_components:
                replacements[component.component_id] = known_components[component]
                replaced.append(i)
                found = True
            else:
                known_components[component] = component.component_id

        if not found:
            break
        for component in mc.components:
            for input in component.inputs:
                if input.component_id in replacements:
                    input.component_id = replacements[input.component_id]

        for i in reversed(replaced):
            mc.components.pop(i)



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class ArithmeticFunction3In(Component):
    """
    Evaluates a mathematical expression with up to 3 input variables and outputs the result.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        function: str,
        x_input: Optional[Input] = None,
        y_input: Optional[Input] = None,
        z_input: Optional[Input] = None,
    ):
        super().__init__(ComponentType.ArithmeticFunction3In, component_id, position)
        self.function: str = function
        self.x_input: Optional[Input] = x_input
        self.y_input: Optional[Input] = y_input
        self.z_input: Optional[Input] = z_input

    @property
    def height(self) -> float:
        return 1.0

    @staticmethod
    def from_xml(element: XMLParserElement) -> ArithmeticFunction3In:
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, _ = Component._basic_in_parsing(obj)
        function: str = obj.attributes.get("e", "")
        return ArithmeticFunction3In(
            component_id,
            position,
            function,
            inputs.get("1"),
            inputs.get("2"),
            inputs.get("3"),
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        return {"e": self.function}, self._pos_in_to_xml(
            self.x_input, self.y_input, self.z_input
        )



from __future__ import annotations

from typing import Optional

from .Input import Input
from .Position import Position
from .Types import NodeMode, SignalType
from .XMLElement import XMLElement
from .XMLParser import XMLParserElement


class NodePosition(XMLElement):
    """
    A Position for a node. This is distinct from :class:`sw_mc_lib.Position.Position` in that it has x, y and z, and
    that they are integers instead of floats. You probably never need to use y.
    """

    def __init__(self, x: int, z: int):
        self.x: int = x
        self.z: int = z

    @staticmethod
    def from_xml(element: XMLParserElement) -> NodePosition:
        assert element.tag == "position", f"Invalid node position {element}"
        return NodePosition(
            int(element.attributes.get("x", "0")),
            int(element.attributes.get("z", "0")),
        )

    def to_xml(self) -> XMLParserElement:
        attributes: dict[str, str] = {}
        if self.x != 0:
            attributes["x"] = str(self.x)
        if self.z != 0:
            attributes["z"] = str(self.z)
        return XMLParserElement("position", attributes)


class Node(XMLElement):
    """
    A Microcontroller node. Interface between Microcontroller and the rest of the Vehicle.
    """

    def __init__(
        self,
        node_id: int,
        component_id: int,
        label: str,
        mode: NodeMode,
        type: SignalType,
        description: str,
        input_position: Optional[NodePosition] = None,
        position: Optional[Position] = None,
        input: Optional[Input] = None,
    ):
        self.node_id: int = node_id
        self.component_id: int = component_id
        self.label: str = label
        self.mode: NodeMode = mode
        self.type: SignalType = type
        self.description: str = description
        self.input_position: Optional[NodePosition] = input_position
        self.position: Optional[Position] = position
        self.input: Optional[Input] = input

    @staticmethod
    def from_xml(element: XMLParserElement) -> Node:
        assert element.tag == "n", f"Invalid node {element}"
        node = element.children[0]
        assert node.tag == "node", f"Invalid node {element}"
        position: Optional[NodePosition] = None
        if node.children:
            position = NodePosition.from_xml(node.children[0])
        return Node(
            int(element.attributes.get("id", "1")),
            int(element.attributes.get("component_id", "1")),
            node.attributes.get("label", ""),
            NodeMode(int(node.attributes.get("mode", "0"))),
            SignalType(int(node.attributes.get("type", "0"))),
            node.attributes.get("description", ""),
            position,
        )

    def get_component_bridge_type(self) -> int:
        """
        Get the component bridge type for this node. This is the type of the mc side of the node,
        while the enum property is the id of the vehicle facing node
        :return: The component bridge type
        """
        # pylint: disable=too-many-return-statements
        match (self.mode, self.type):
            case (NodeMode.Input, SignalType.Boolean):
                return 0
            case (NodeMode.Output, SignalType.Boolean):
                return 1
            case (NodeMode.Input, SignalType.Number):
                return 2
            case (NodeMode.Output, SignalType.Number):
                return 3
            case (NodeMode.Input, SignalType.Composite):
                return 4
            case (NodeMode.Output, SignalType.Composite):
                return 5
            case (NodeMode.Input, SignalType.Video):
                return 6
            case (NodeMode.Output, SignalType.Video):
                return 7
            case (NodeMode.Input, SignalType.Audio):
                return 8
            case (NodeMode.Output, SignalType.Audio):
                return 9
            case _:
                raise ValueError(
                    f"Invalid node mode/type combination {self.mode}/{self.type}"
                )

    def add_component_bridge(self, element: XMLParserElement) -> None:
        """Add component bridge data to this node from XML (mc side of the node)"""
        assert element.tag == "c", f"Invalid node {element}"
        assert element.attributes.get("type", "0") == str(
            self.get_component_bridge_type()
        ), f"Invalid node {element}"
        node = element.children[0]
        assert node.tag == "object", f"Invalid node {element}"
        assert node.attributes.get("id", "0") == str(
            self.component_id
        ), f"Invalid node {element}"
        for child in node.children:
            if child.tag == "in1":
                self.input = Input.from_xml(child)
            elif child.tag == "pos":
                self.position = Position.from_xml(child)

    def to_component_bridge(self) -> XMLParserElement:
        """Convert this Node to a XML component bridge entity (mc side of the node)"""
        children = []
        if self.input:
            children.append(self.input.to_xml())
        if self.position:
            children.append(self.position.to_xml())
        object_ = XMLParserElement("object", {"id": str(self.component_id)}, children)
        return XMLParserElement(
            "c", {"type": str(self.get_component_bridge_type())}, [object_]
        )

    def to_component_bridge_state(self, index: int) -> XMLParserElement:
        """Convert this Node to a XML component bridge state entity (mc side of the node)"""
        children = []
        if self.input:
            self.input.index = "1"
            children.append(self.input.to_xml())
        if self.position:
            children.append(self.position.to_xml())
        return XMLParserElement(f"c{index}", {"id": str(self.component_id)}, children)

    def to_xml(self) -> XMLParserElement:
        node_attributes: dict[str, str] = {
            "label": self.label,
            "mode": str(self.mode.value),
            "type": str(self.type.value),
            "description": self.description,
        }
        node_element: XMLParserElement = XMLParserElement("node", node_attributes)
        if self.input_position:
            node_element.children.append(self.input_position.to_xml())
        return XMLParserElement(
            "n",
            {"id": str(self.node_id), "component_id": str(self.component_id)},
            [node_element],
        )

    def __hash__(self) -> int:
        return hash((self.node_id, self.component_id, self.label, self.mode, self.type))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Node):
            return NotImplemented
        return (
            self.node_id == other.node_id
            and self.component_id == other.component_id
            and self.label == other.label
            and self.mode == other.mode
            and self.type == other.type
            and self.description == other.description
            and self.input_position == other.input_position
            and self.position == other.position
            and self.input == other.input
        )

    def clone(self) -> Node:
        """Clone this node by converting to and from XML"""
        return self.from_xml(self.to_xml())



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.NumberProperty import NumberInput, NumberProperty
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class PIDController(Component):
    """
    A basic PID controller. The proportional, integral and derivative gains can be set in the property panel.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        setpoint_input: Optional[Input] = None,
        process_variable_input: Optional[Input] = None,
        active_input: Optional[Input] = None,
        proportional_property: NumberInput = None,
        integral_property: NumberInput = None,
        derivative_property: NumberInput = None,
    ):
        super().__init__(ComponentType.PIDController, component_id, position)
        self.setpoint_input: Optional[Input] = setpoint_input
        self.process_variable_input: Optional[Input] = process_variable_input
        self.active_input: Optional[Input] = active_input
        self.proportional_property: NumberProperty = NumberProperty.from_input(
            proportional_property, "kp"
        )
        self.integral_property: NumberProperty = NumberProperty.from_input(
            integral_property, "ki"
        )
        self.derivative_property: NumberProperty = NumberProperty.from_input(
            derivative_property, "kd"
        )

    @property
    def height(self) -> float:
        return 1.0

    @staticmethod
    def from_xml(element: XMLParserElement) -> PIDController:
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, properties = Component._basic_in_parsing(obj)
        return PIDController(
            component_id,
            position,
            inputs.get("1"),
            inputs.get("2"),
            inputs.get("3"),
            properties.get("kp"),
            properties.get("ki"),
            properties.get("kd"),
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        return {}, self._pos_in_to_xml(
            self.setpoint_input,
            self.process_variable_input,
            self.active_input,
            properties={
                "kp": self.proportional_property,
                "ki": self.integral_property,
                "kd": self.derivative_property,
            },
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from inspect import Signature, signature
from typing import Any

from .util import generic_str
from .XMLParser import XMLParserElement


class XMLElement(ABC):
    """
    An element that can be deserialized from and serialized to an XML element
    """

    @abstractmethod
    def to_xml(self) -> XMLParserElement:
        """
        Serialize the element to an XML element.

        :return: The XML element
        """

    @staticmethod
    @abstractmethod
    def from_xml(element: XMLParserElement) -> XMLElement:
        """
        Deserialize the element from an XML element.

        :param element: XML element
        :return: Deserialized element
        """

    def __repr__(self) -> str:
        return generic_str(self)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        self_signature: Signature = signature(self.__init__)  # type: ignore
        return all(
            hasattr(other, name) and getattr(self, name) == getattr(other, name)
            for name in self_signature.parameters.keys()
        )



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.NumberProperty import NumberInput, NumberProperty
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class MemoryRegister(Component):
    """
    "Remembers the input value when receiving a signal to the Set node. When the Reset node receives a signal,
    the stored number is cleared to a value that can be customised in the properties panel.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        reset_property: NumberInput = None,
        set_input: Optional[Input] = None,
        reset_input: Optional[Input] = None,
        number_to_store_input: Optional[Input] = None,
    ):
        super().__init__(ComponentType.MemoryRegister, component_id, position)
        self.set_input: Optional[Input] = set_input
        self.reset_input: Optional[Input] = reset_input
        self.number_to_store_input: Optional[Input] = number_to_store_input
        self.reset_property: NumberProperty = NumberProperty.from_input(
            reset_property, "r"
        )

    @property
    def height(self) -> float:
        return 1.0

    @staticmethod
    def from_xml(element: XMLParserElement) -> MemoryRegister:
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, properties = Component._basic_in_parsing(obj)
        return MemoryRegister(
            component_id,
            position,
            properties.get("r"),
            inputs.get("1"),
            inputs.get("2"),
            inputs.get("3"),
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        return {}, self._pos_in_to_xml(
            self.set_input,
            self.reset_input,
            self.number_to_store_input,
            properties={"r": self.reset_property},
        )



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class JKFlipFlop(Component):
    """
    An JK flip flop that can be set and reset using two on/off inputs.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        set_input: Optional[Input] = None,
        reset_input: Optional[Input] = None,
    ):
        super().__init__(ComponentType.JKFlipFlop, component_id, position)
        self.set_input: Optional[Input] = set_input
        self.reset_input: Optional[Input] = reset_input

    @property
    def height(self) -> float:
        return 0.75

    @staticmethod
    def from_xml(element: XMLParserElement) -> JKFlipFlop:
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, _ = Component._basic_in_parsing(obj)
        return JKFlipFlop(component_id, position, inputs.get("1"), inputs.get("2"))

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        return {}, self._pos_in_to_xml(self.set_input, self.reset_input)



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType, TimerUnit
from sw_mc_lib.XMLParser import XMLParserElement


class TimerRTF(Component):
    """
    Variable input timer. Outputs an on signal when the timer is less than its duration.
    The timer will not reset until it is signalled.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        unit_property: TimerUnit = TimerUnit.Seconds,
        timer_enable_input: Optional[Input] = None,
        duration_input: Optional[Input] = None,
        reset_input: Optional[Input] = None,
    ):
        super().__init__(ComponentType.TimerRTF, component_id, position)
        self.timer_enable_input: Optional[Input] = timer_enable_input
        self.duration_input: Optional[Input] = duration_input
        self.reset_input: Optional[Input] = reset_input
        self.unit_property: TimerUnit = unit_property

    @property
    def height(self) -> float:
        return 1.0

    @staticmethod
    def from_xml(element: XMLParserElement) -> TimerRTF:
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, _ = Component._basic_in_parsing(obj)
        unit_property: TimerUnit = TimerUnit(int(obj.attributes.get("u", "0")))
        return TimerRTF(
            component_id,
            position,
            unit_property,
            inputs.get("1"),
            inputs.get("2"),
            inputs.get("3"),
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        return {"u": str(self.unit_property.value)}, self._pos_in_to_xml(
            self.timer_enable_input, self.duration_input, self.reset_input
        )



from __future__ import annotations

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.NumberProperty import NumberInput, NumberProperty
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class PropertyNumber(Component):
    """
    Adds a custom number input that will be seen on the microcontroller's property panel when placed on a vehicle.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        name: str = "number",
        value_property: NumberInput = None,
    ):
        super().__init__(ComponentType.PropertyNumber, component_id, position)
        self.name: str = name
        self.value_property: NumberProperty = NumberProperty.from_input(
            value_property, "v"
        )

    @property
    def height(self) -> float:
        return 0.5

    @staticmethod
    def from_xml(element: XMLParserElement) -> PropertyNumber:
        obj: XMLParserElement = element.children[0]
        component_id, position, _, properties = Component._basic_in_parsing(obj)
        name: str = obj.attributes.get("n", "number")
        return PropertyNumber(component_id, position, name, properties.get("v"))

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        children: list[XMLParserElement] = self._pos_in_to_xml(
            properties={"v": self.value_property}
        )
        return {}, children



from __future__ import annotations

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.NumberProperty import NumberInput, NumberProperty
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class ConstantNumber(Component):
    """
    Outputs a constant number that is set on the properties panel.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        value_property: NumberInput = None,
    ):
        super().__init__(ComponentType.ConstantNumber, component_id, position)
        self.value_property: NumberProperty = NumberProperty.from_input(
            value_property, "n"
        )

    @property
    def height(self) -> float:
        return 0.5

    @staticmethod
    def from_xml(element: XMLParserElement) -> ConstantNumber:
        obj: XMLParserElement = element.children[0]
        component_id, position, _, properties = Component._basic_in_parsing(obj)
        return ConstantNumber(component_id, position, properties.get("n"))

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        return {}, self._pos_in_to_xml(properties={"n": self.value_property})
